import { asyncIter } from "./extended_async_iterable";
let counter = 0;
/*
 * wrap an `AsyncIterable<Uint8Array>` so that it can provide byte-level read
 * operations.
 */
export class ByteReader {
    iterable;
    iter;
    saved = [];
    size = 0;
    ended = false;
    // track how many bytes were read
    bytesRead = 0;
    id = ++counter;
    constructor(iterable) {
        this.iterable = iterable;
        this.iter = iterable[Symbol.asyncIterator]();
    }
    // turn this back into an async iterator.
    toIterable() {
        const buffer = this.remainder();
        const iter = this.iter;
        // poison the byte reader. it can't work anymore.
        delete this.iter;
        return asyncIter(async function* () {
            if (buffer)
                yield buffer;
            yield* asyncIter(iter);
        }());
    }
    /*
     * read `size` bytes, waiting for data to arrive if necessary.
     * if the stream ends before enough data is received, it will return as
     * much as it can. if the stream has already ended, it will return
     * `undefined`.
     */
    async read(size) {
        await this.fillTo(size);
        if (this.saved.length == 0)
            return undefined;
        return this.splitOff(size);
    }
    // returns true if anything was added.
    async fill() {
        if (this.ended)
            return false;
        const item = await this.iter.next();
        if (item.done) {
            this.ended = true;
            return false;
        }
        this.saved.push(item.value);
        this.size += item.value.length;
        return true;
    }
    async fillTo(size) {
        while (this.size < size) {
            if (!await this.fill())
                return;
        }
    }
    // this works even if size > this.size
    splitOff(size) {
        const total = concatBuffers(this.saved);
        const rv = total.slice(0, size);
        if (size >= total.length) {
            this.saved = [];
            this.size = 0;
        }
        else {
            this.saved = [total.slice(size)];
            this.size = this.saved[0].length;
        }
        this.bytesRead += rv.length;
        return rv;
    }
    find(byte) {
        let total = 0;
        for (let i = 0; i < this.saved.length; i++) {
            const n = this.saved[i].indexOf(byte);
            if (n >= 0)
                return total + n;
            total += this.saved[i].length;
        }
        return undefined;
    }
    /*
     * read and buffer data from this stream until a `test` returns an index
     * indicating a successful match. if no match is found before the stream
     * ends, it returns `undefined`.
     * `test` is called with all currently buffered data, each time new data
     * arrives. it should return -1 on an unsuccessful match. on success, it
     * should return the index right after the end of the match (as if you were
     * about to pass it to `slice`).
     */
    async readUntilMatch(test) {
        while (true) {
            const buffer = concatBuffers(this.saved);
            const n = test(buffer);
            if (n >= 0) {
                this.saved = [buffer.slice(n)];
                this.size = this.saved[0].length;
                this.bytesRead += n;
                return buffer.slice(0, n);
            }
            if (!await this.fill())
                return undefined;
        }
    }
    /*
     * read and buffer data from this stream until a specific byte is seen.
     * once it's seen, all buffered data up to & including the requested byte
     * is returned. if the byte is never seen before the stream ends, it
     * returns `undefined`.
     */
    async readUntil(byte) {
        const n = this.find(byte);
        if (n !== undefined)
            return this.splitOff(n + 1);
        if (this.ended)
            return undefined;
        while (true) {
            const oldTotal = this.size;
            if (!await this.fill())
                return undefined;
            const n = this.saved[this.saved.length - 1].indexOf(byte);
            if (n >= 0)
                return this.splitOff(oldTotal + n + 1);
        }
    }
    /*
     * generate a stream which passes through data from this stream until `test`
     * returns an index indicating a successful match, or this stream ends.
     * if a match was found, after the new stream ends, this (original)
     * stream will resume at the start of the match. otherwise, the original
     * stream will end too.
     *
     * `test` is called on each buffer as it arrives from the underlying
     * stream. it should return -1 on an unsuccessful match. on success, it
     * should return the index of the start of the match.
     *
     * if `sizeHint` is greater than 1, each subsequent call to `test` will
     * have `sizeHint - 1` bytes repeated from the end of the previous buffer,
     * so that no `sizeHint`-length span of the stream will be broken up across
     * calls to `test`. (this is used by `streamUntilBuffer`.)
     */
    streamUntilMatch(test, sizeHint = 1) {
        const keepOverlap = sizeHint > 1 ? sizeHint - 1 : 0;
        // absorb our entire buffer until this is finished.
        let buffer = concatBuffers(this.saved);
        this.saved = [];
        this.size = 0;
        const self = this;
        return asyncIter(async function* () {
            while (true) {
                const n = test(buffer);
                if (n >= 0) {
                    if (n > 0)
                        yield buffer.subarray(0, n);
                    self.saved = [buffer.subarray(n)];
                    self.size = self.saved[0].length;
                    self.bytesRead += n;
                    return;
                }
                else if (buffer.length > keepOverlap) {
                    yield buffer.subarray(0, buffer.length - keepOverlap);
                    buffer = buffer.subarray(buffer.length - keepOverlap);
                    self.bytesRead += buffer.length - keepOverlap;
                }
                const item = await self.iter.next();
                if (item.done) {
                    if (buffer.length > 0)
                        yield buffer;
                    self.ended = true;
                    return;
                }
                buffer = concatBuffers([buffer, item.value]);
            }
        }());
    }
    /*
     * generate a stream which passes through data from this stream until an
     * exact match for `match` is found, or this stream ends. if a match was
     * found, after the new stream ends, this (original) stream will resume at
     * the start of the match. otherwise, the original stream will end too.
     */
    streamUntilBuffer(match) {
        return this.streamUntilMatch(b => {
            // optimize:
            let i = 0;
            const max = b.length - match.length;
            while (i >= 0 && i < max) {
                i = b.indexOf(match[0], i);
                if (i >= 0) {
                    if (i < max && Buffer.from(b.subarray(i, i + match.length)).compare(match) == 0)
                        return i;
                    i++;
                }
            }
            return -1;
        }, match.length);
    }
    /*
     * "unshift" a buffer back onto the head of the stream.
     */
    unread(buffer) {
        this.saved.unshift(buffer);
        this.size += buffer.length;
        this.bytesRead -= buffer.length;
    }
    /*
     * return all currently buffered data without waiting for more, and clear
     * the buffers. if nothing is buffered, it returns `undefined`.
     */
    remainder() {
        if (this.size == 0)
            return undefined;
        const ret = concatBuffers(this.saved);
        this.saved = [];
        this.size = 0;
        this.bytesRead += ret.length;
        return ret;
    }
    toString() {
        return `ByteReader[${this.id}](read=${this.bytesRead}, buffered=${this.size}, ${this.iterable.toString()})`;
    }
}
/*
 * Turn any existing `Uint8Array` stream into a `ByteReader` with useful methods.
 */
export function byteReader(wrapped, getDebugName) {
    if (wrapped instanceof ByteReader)
        return wrapped;
    return new ByteReader(asyncIter(wrapped, getDebugName));
}
export function concatBuffers(buffers) {
    // it kiiiiiiiinda sucks that we have to spell this out.
    const len = buffers.reduce((sum, b) => sum + b.length, 0);
    const ret = new Uint8Array(len);
    let offset = 0;
    for (const b of buffers) {
        ret.set(b, offset);
        offset += b.length;
    }
    return ret;
}
//# sourceMappingURL=byte_reader.js.map