import { Readable } from "node:stream";
import { Gate } from "./gate";
let counter = 0;
/*
 * wrapper for AsyncIterable that has basic functional operations on it.
 */
export class ExtendedAsyncIterable {
    wrapped;
    id = ++counter;
    getDebugName = () => this.wrapped.toString();
    cachedIterator;
    constructor(wrapped) {
        this.wrapped = wrapped;
        // pass
    }
    [Symbol.asyncIterator]() {
        // if the underlying iterable is restartable, we should be too:
        this.cachedIterator = this.wrapped[Symbol.asyncIterator]();
        return this;
    }
    next() {
        this.cachedIterator = this.cachedIterator ?? this.wrapped[Symbol.asyncIterator]();
        return this.cachedIterator.next();
    }
    map(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped)
                yield f(item);
        }());
    }
    flatMap(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped)
                yield* f(item);
        }());
    }
    filter(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped)
                if (f(item))
                    yield item;
        }());
    }
    filterMap(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped) {
                const rv = f(item);
                if (rv !== undefined)
                    yield rv;
            }
        }());
    }
    async find(f) {
        for await (const item of this.wrapped) {
            if (f(item))
                return item;
        }
        return undefined;
    }
    async some(f) {
        return (await this.find(f)) !== undefined;
    }
    async every(f) {
        return (await this.find(x => !f(x))) === undefined;
    }
    async reduce(f, start) {
        let rv = start;
        for await (const x of this.wrapped)
            rv = f(rv, x);
        return rv;
    }
    async collect() {
        return this.reduce((sum, x) => { sum.push(x); return sum; }, []);
    }
    chain(iter) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped)
                yield item;
            for await (const item of iter)
                yield item;
        }());
    }
    static chainAll(iters) {
        return asyncIter(async function* () {
            for (const iter of iters) {
                for await (const item of iter)
                    yield item;
            }
        }());
    }
    zip(iter) {
        const iter1 = this.wrapped[Symbol.asyncIterator]();
        const iter2 = iter[Symbol.asyncIterator]();
        return asyncIter(async function* () {
            while (true) {
                const r1 = await iter1.next();
                const r2 = await iter2.next();
                if (r1.done || r2.done)
                    return;
                yield [r1.value, r2.value];
            }
        }());
    }
    merge(...iterables) {
        return ExtendedAsyncIterable.mergeAll([this, ...iterables]);
    }
    static mergeAll(iterables) {
        const iters = iterables.map(iter => iter[Symbol.asyncIterator]());
        const next = async (iter) => [iter, await iter.next()];
        const queue = new Map(iters.map(iter => [iter, next(iter)]));
        return asyncIter(async function* () {
            while (queue.size > 0) {
                const [iter, result] = await Promise.race(queue.values());
                if (result.done) {
                    queue.delete(iter);
                }
                else {
                    queue.set(iter, next(iter));
                    yield result.value;
                }
            }
        }());
    }
    enumerate() {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            let i = 0;
            for await (const item of wrapped) {
                yield [i, item];
                i++;
            }
        }());
    }
    // return 2 iterators:
    //   - the first contains items as long as f(item) is false
    //   - the second contains the first item where f(item) is true, and all
    //     remaining items after that
    splitWhen(f) {
        const iter = this.wrapped[Symbol.asyncIterator]();
        let unblock;
        const blocked = new Promise(resolve => {
            unblock = resolve;
        });
        async function* iter1() {
            while (true) {
                const r = await iter.next();
                if (r.done)
                    break;
                if (f(r.value)) {
                    unblock(r.value);
                    return;
                }
                yield r.value;
            }
            unblock(undefined);
        }
        async function* iter2() {
            const item = await blocked;
            if (item === undefined)
                return;
            yield item;
            while (true) {
                const r = await iter.next();
                if (r.done)
                    return;
                yield r.value;
            }
        }
        return [asyncIter(iter1()), asyncIter(iter2())];
    }
    tee(count = 2) {
        const iter = this.wrapped[Symbol.asyncIterator]();
        const queues = [...Array(count).keys()].map(i => []);
        const gate = new Gate(async () => {
            const r = await iter.next();
            for (const q of queues)
                q.push(r);
        });
        return queues.map(q => asyncIterFromQ(q, () => gate.wait()));
    }
    partition(f) {
        const iter = this.wrapped[Symbol.asyncIterator]();
        const queues = [[], []];
        const gate = new Gate(async () => {
            const r = await iter.next();
            if (r.done) {
                queues[0].push(r);
                queues[1].push(r);
            }
            else {
                queues[f(r.value) ? 0 : 1].push(r);
            }
        });
        return [asyncIterFromQ(queues[0], () => gate.wait()), asyncIterFromQ(queues[1], () => gate.wait())];
    }
    takeWhile(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped) {
                if (!f(item))
                    return;
                yield item;
            }
        }());
    }
    take(n) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            let remaining = n;
            for await (const item of wrapped) {
                yield item;
                remaining--;
                if (remaining == 0)
                    return;
            }
        }());
    }
    takeFor(msec) {
        return this.takeUntil(Date.now() + msec);
    }
    takeUntil(deadline) {
        const iter = this.wrapped[Symbol.asyncIterator]();
        return asyncIter(async function* () {
            while (true) {
                const now = Date.now();
                if (now >= deadline)
                    return;
                const nextPromise = iter.next();
                const result = await Promise.race([
                    nextPromise,
                    new Promise(resolve => setTimeout(resolve, deadline - now)).then(() => undefined)
                ]);
                // ran out of time? throw it away. (js doesn't have cancel)
                if (result === undefined || result.done)
                    return;
                yield result.value;
            }
        }());
    }
    dropWhile(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            let dropping = true;
            for await (const item of wrapped) {
                if (dropping && f(item))
                    continue;
                dropping = false;
                yield item;
            }
        }());
    }
    drop(n) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            let remaining = n;
            for await (const item of wrapped) {
                if (remaining > 0) {
                    remaining--;
                }
                else {
                    yield item;
                }
            }
        }());
    }
    after(f) {
        const wrapped = this.wrapped;
        return asyncIter(async function* () {
            for await (const item of wrapped)
                yield item;
            await f();
        }(), () => `after(${this})`);
    }
    withPromiseAfter() {
        let resolve = () => 0;
        const done = new Promise(r => {
            resolve = r;
        });
        return [this.after(async () => resolve()), done];
    }
    toString() {
        return `ExtendedAsyncIterable[${this.id}](${this.getDebugName()})`;
    }
    toNodeStream(options = {}) {
        const iter = this.wrapped[Symbol.asyncIterator]();
        let waiting = false;
        const r = new Readable(Object.assign({}, options, {
            async read(size) {
                // ignore redundant read() calls when we're busy filling the pipe
                if (waiting)
                    return;
                waiting = true;
                try {
                    // keep reading until the outbound buffer says to stop
                    while (true) {
                        const item = await iter.next();
                        if (item.done ?? false) {
                            r.push(null);
                            break;
                        }
                        else {
                            if (!r.push(item.value))
                                break;
                        }
                    }
                }
                catch (error) {
                    if (error instanceof Error) {
                        r.destroy(error);
                    }
                    else {
                        throw error;
                    }
                }
                waiting = false;
            }
        }));
        return r;
    }
}
// small wrapper to allow an iterator to be iterable.
class AsyncIterableIterator {
    __iter;
    constructor(__iter) {
        this.__iter = __iter;
    }
    [Symbol.asyncIterator]() { return this.__iter; }
    toString() { return this.__iter.toString(); }
}
/*
 * Turn an `AsyncIterable` or `AsyncIterator` into an `ExtendedAsyncIterable`
 * that has functional methods on it. Will try to add as few wrappers as
 * possible, and will pass existing `ExtendedAsyncIterable`s through unharmed.
 */
export function asyncIter(iter, getDebugName) {
    // already ExtendedAsyncIterable?
    if (iter instanceof ExtendedAsyncIterable) {
        if (getDebugName)
            iter.getDebugName = getDebugName;
        return iter;
    }
    // AsyncIterable?
    if (iter[Symbol.asyncIterator]) {
        const i = new ExtendedAsyncIterable(iter);
        if (getDebugName)
            i.getDebugName = getDebugName;
        return i;
    }
    // Iterable?
    if (iter[Symbol.iterator]) {
        return asyncIter(async function* () {
            for (const item of iter)
                yield item;
        }(), getDebugName || (() => cheapInspect(iter)));
    }
    // AsyncIterator
    const i = new ExtendedAsyncIterable(new AsyncIterableIterator(iter));
    if (getDebugName)
        i.getDebugName = getDebugName;
    return i;
}
function cheapInspect(x) {
    if (x["inspect"])
        return x["inspect"]();
    if (x[Symbol.iterator]) {
        // it's an Iterable
        const iter = x;
        if (Array.isArray(iter)) {
            const a = iter;
            if (a.length == 0)
                return "[]";
            return "[ " + a.map(cheapInspect).join(", ") + " ]";
        }
        return "Iterable(...)";
    }
    return x.constructor.name + "(...)";
}
// factored out from tee & partition: given a queue of saved-up items, and
// a function that waits for more to be available, generate an asyncIter.
function asyncIterFromQ(q, next) {
    return asyncIter(async function* () {
        while (true) {
            let result = q.shift();
            if (result === undefined) {
                await next();
            }
            else {
                if (result.done)
                    return;
                yield result.value;
            }
        }
    }());
}
//# sourceMappingURL=extended_async_iterable.js.map